@@ -29,14 +29,16 @@ def setup_m3_bandset_metadata(file_path):
     metadata["IX"] = np.array(range(len(metadata["WAVELENGTH"])))
     # 'channel' or 'channel #' in mission nomenclature
     metadata["BAND"] = metadata["IX"] + 1
-    metadata['PATH'] = file_path
+    metadata["PATH"] = file_path
     return metadata
 
 
 class M3BandSet(BandSet):
     def __init__(self, file_path, rois=None, threads=None):
         metadata = setup_m3_bandset_metadata(file_path)
-        load_method = partial(pdr_load, preserve_constants=-999.0)
+        load_method = partial(
+            pdr_load, preserve_constants=[-999.0], object_name='PRIMARY'
+        )
         super().__init__(
             metadata=metadata,
             load_method=load_method,
